#ifndef PQCLEAN_MLDSA44_AVX2_CONSTS_H
#define PQCLEAN_MLDSA44_AVX2_CONSTS_H
#include "align.h"
#include "cdecl.h"


typedef ALIGNED_INT32(624) qdata_t;
extern const qdata_t PQCLEAN_MLDSA44_AVX2_qdata;

#endif
